//
// Compare performance of converting a string to its floating point value
// using the following methods:
// 
// Timing results (smaller is faster):
//      
// WIN32 Release
//  Time           Method  Computed sum 
// ------      ---------- ----------------------
//  0.156            AToF sum=20246804246763.414
//  0.343            atof sum=20246804246763.414
//  0.328          strtod sum=20246804246763.414
//  0.561           scanf sum=20246804246763.414
//  3.510    stringStream sum=20246804246763.414
//  3.167       strStream sum=20246804246763.414
// 
//
// X64 Release
//  Time           Method  Computed sum 
// ------      ---------- ----------------------
//  0.187            AToF sum=20246804246763.414
//  0.266            atof sum=20246804246763.414
//  0.328          strtod sum=20246804246763.414
//  0.500           scanf sum=20246804246763.414
//  2.340    stringStream sum=20246804246763.414
//  2.091       strStream sum=20246804246763.414
// 
// Conclusion:  AToF is fastest.
//
//
// Dennis Lang 2009
// http://home.comcast.net/~lang.dennis/
// 
 

#include <vector>
#include <string>
#include <iostream>
#include <iomanip>
#include <sstream>
#include <strstream>
#include <assert.h>


#include <limits.h>
#include <math.h>
#include <ctype.h>
#define LOGHUGE 39                        /* 10^39 --> Largest power */

//  http://www.bsdlover.cn/study/UnixTree/V7/usr/src/libc/gen/atof.c.html
double AToF(register const char *p)
{
	register char c;
	double fl, flexp, exp5;
	double big = 72057594037927936.;  /*2^56*/
	double ldexp();
	int nd;
	register int eexp, exp, neg, negexp, bexp;

	neg = 1;
	while((c = *p++) == ' ')
		;
	if (c == '-')
		neg = -1;
	else if (c=='+')
		;
	else
		--p;

	exp = 0;
	fl = 0;
	nd = 0;
	while ((c = *p++), isdigit(c)) 
    {
		if (fl<big)
			fl = 10*fl + (c-'0');
		else
			exp++;
		nd++;
	}

	if (c == '.') 
    {
		while ((c = *p++), isdigit(c)) 
        {
			if (fl<big) {
				fl = 10*fl + (c-'0');
				exp--;
			}
		nd++;
		}
	}

	negexp = 1;
	eexp = 0;
	if ((c == 'E') || (c == 'e')) 
    {
		if ((c= *p++) == '+')
			;
		else if (c=='-')
			negexp = -1;
		else
			--p;

		while ((c = *p++), isdigit(c)) 
        {
			eexp = 10*eexp+(c-'0');
		}
		if (negexp<0)
			eexp = -eexp;
		exp = exp + eexp;
	}

	negexp = 1;
	if (exp<0) {
		negexp = -1;
		exp = -exp;
	}


	if((nd+exp*negexp) < -LOGHUGE)
    {
		fl = 0;
		exp = 0;
	}
	flexp = 1;
	exp5 = 5;
	bexp = exp;

	for (;;) 
    {
		if (exp&01)
			flexp *= exp5;
		exp >>= 1;
		if (exp==0)
			break;
		exp5 *= exp5;
	}
	if (negexp<0)
		fl /= flexp;
	else
		fl *= flexp;
    fl = ::ldexp(fl, negexp*bexp);
	if (neg<0)
		fl = -fl;
	return(fl);
}

typedef std::vector<std::string> Words;

double TestAToF(Words& words)
{
    double sum = 0;

    for (size_t n = 0; n != words.size(); n++)
        sum += AToF(words[n].c_str());

    return sum;
}


double TestAtof(Words& words)
{
    double sum = 0;

    for (size_t n = 0; n != words.size(); n++)
        sum += atof(words[n].c_str());

    return sum;
}

double TestStrtod(Words& words)
{
    double sum = 0;

    for (size_t n = 0; n != words.size(); n++)
        sum += strtod(words[n].c_str(), NULL);

    return sum;
}


double TestScanf(Words& words)
{
    double sum = 0;
    float fnum;

    for (size_t n = 0; n != words.size(); n++)
    {
        sscanf(words[n].c_str(), "%f", &fnum);
        sum += fnum;
    }

    return sum;
}


double TestStringStream(Words& words)
{
    double sum = 0;
    double fnum;

    for (size_t n = 0; n != words.size(); n++)
    {
        std::istringstream in(words[n].c_str());
        in >> fnum;
        // assert(in.good());
        sum += fnum;
    }

    return sum;
}


double TestStrStream(Words& words)
{
    double sum = 0;
    double fnum;

    for (size_t n = 0; n != words.size(); n++)
    {
        std::istrstream in(words[n].c_str());
        in >> fnum;
        // assert(in.good());
        sum += fnum;
    }

    return sum;
}


void MakeFloatWords(size_t nWords, Words& words)
{
    char sbuf[256];
    float values[] = { -123.456e-6f, -0, 0, +0, 123.456f, 123.456e6f };

    while (nWords-- != 0)
    {
        float fnum = values[nWords % sizeof(values)/sizeof(values[0])];
        sprintf_s(sbuf, sizeof(sbuf), "%f", fnum);
        words.push_back(std::string(sbuf));
    }
}

#include <windows.h>
typedef double (*fptr)(Words&);

double TimeIt(const char* what, fptr func, Words& w)
{
    DWORD start = GetTickCount();
    double d = 0;
    for (int t=0; t < 1000; t++)
        d += func(w);
    DWORD end   = GetTickCount();
    std::cout << std::fixed << std::setprecision(3) << std::setw(8) << (end - start) / 1000.0
        << " " << std::setw(15) << what << " sum=" << d << std::endl;

    return (end - start) / 1000.0;
}

int main(int argc, char* argv[])
{
    Words floatWords;

    MakeFloatWords(1000, floatWords);
    TimeIt("AToF", TestAToF, floatWords);
    TimeIt("atof", TestAtof, floatWords);
    TimeIt("strtod", TestStrtod, floatWords);
    TimeIt("scanf", TestScanf, floatWords);
    TimeIt("stringStream", TestStringStream, floatWords);
    TimeIt("strStream", TestStrStream, floatWords);

	return 0;
}

