//
// Compare performance of converting a float to a string
// using the following methods:
// 
// Timing results (smaller is faster):
// WIN32 Release
//
//    Time          Method  Presentation
// -------    ------------ -----------------
//   0.827            fcvt -12345.600
//   1.669            gcvt -12345.59961
//   1.108         sprintf -12345.600
//   3.806       strStream -12345.600
//   4.009    stringStream -12345.600
// 
//
//  X64 Release
//    Time          Method  Presentation
// -------    ------------ -----------------
//   0.671            fcvt -12345.600
//   1.326            gcvt -12345.59961
//   0.905         sprintf -12345.600
//   2.793       strStream -12345.600
//   2.792    stringStream -12345.600
//
//  Conclusion:  fcvt is fastest
//
//
// Dennis Lang 2009
// http://home.comcast.net/~lang.dennis/
// 
#include <vector>
#include <string>
#include <iostream>
#include <iomanip>
#include <sstream>
#include <strstream>
#include <assert.h>


#include <limits.h>
#include <math.h>
#include <ctype.h>
#include <stdlib.h> 

const int sWidth = 10;
const int sPrecision = 3;
const char sFmt[] = "%10.3f";

typedef std::vector<float> Words;

void MakeFloatWords(size_t nWords, Words& words)
{
    float values[] = { -123.456e-6f, -0.1f, 0, 0.1f, 123.456f, 123.456e6f, 1.234f, -1.234f, 123.456e-2, -123.456e2 };
    const int nValues = sizeof(values)/sizeof(values[0]);

    while (nWords-- != 0)
    {
        float fnum = values[nWords % nValues];
        words.push_back(fnum);
    }
}

#include <windows.h>    // Need for GetTickCount()
typedef const char* (*fptr)(Words&, long dummyValue);

double TimeIt(const char* what, fptr func, Words& w, long dummyValue)
{
    DWORD start = GetTickCount();
    unsigned long sumPtr = 0;
    std::string resultStr;
    for (int t=0; t < 1000; t++)
    {
        const char* str = func(w, dummyValue);
        sumPtr += (unsigned long)str;  // keep compiler from optimizing out loop.
        resultStr = str;
    }

    DWORD end   = GetTickCount();
    if (sumPtr != 0)
    {
        std::cout << std::fixed << std::setprecision(3) << std::setw(8) << (end - start) / 1000.0
            << " " << std::setw(15) << what << " " << resultStr << std::endl;
    }

    return (end - start) / 1000.0;
}

const char* TestSprintf(Words& words, long dummyValue)
{
    static std::string str;
    char strBuf[20];
    int sign, decPos;
    for (size_t n = words.size()-1; n < words.size(); n--)
    {
        _snprintf(strBuf, sizeof(strBuf), sFmt, words[n]);
        str = strBuf;
        if (dummyValue == (long)str.c_str())
            break;
    }

    return str.c_str();  
}

const char* TestFcvt(Words& words, long dummyValue)
{
    static std::string str;
    const int sDecShift = 20;
    char strBuf[20 + sDecShift];
    char* strDst = strBuf + sDecShift;
    const int strSize = sizeof(strBuf) - sDecShift;
    int sign, decPos;

    for (size_t n = words.size()-1; n < words.size(); n--)
    {
        char* strDst = strBuf + sDecShift;
        float fValue = words[n];

        _fcvt_s(strDst, strSize, fValue, sPrecision, &decPos, &sign);
        if (decPos > 0)
        {
            char* decStr = strDst + decPos;
            memmove(decStr+1, decStr, strSize - decPos - 1);
            *decStr = '.';
        }
        else 
        {
            if (decPos < 0)
            {
                memset(strDst + decPos, '0', -decPos);
                strDst += decPos;
            }
            *--strDst = '.';
            *--strDst = '0';
        }
        
        if (sign != 0) 
            *--strDst = '-';
        str = strDst;
        if (dummyValue == (long)str.c_str())
            break;
    }

    return str.c_str();
}

const char* TestGcvt(Words& words, long dummyValue)
{
    static std::string str;
    char strBuf[20];
    int sign, decPos;

    for (size_t n = words.size()-1; n < words.size(); n--)
    {
        _gcvt_s(strBuf, sizeof(strBuf), words[n], sWidth);
        str = strBuf;
        if (dummyValue == (long)str.c_str())
            break;
    }

    return str.c_str();
}

const char* TestEcvt(Words& words, long dummyValue)
{
    static std::string str;
    char strBuf[20];
    int sign, decPos;

    for (size_t n = words.size()-1; n < words.size(); n--)
    {
        _ecvt_s(strBuf+1, sizeof(strBuf)-1, words[n], sWidth - sPrecision, &decPos, &sign);
        strBuf[0] = (sign != 0) ? '-' : ' ';
        str = strBuf;
        if (dummyValue == (long)str.c_str())
            break;
    }

    return str.c_str();
}

const char* TestStrStream(Words& words, long dummyValue)
{
    static std::string str;
    for (size_t n = words.size()-1; n < words.size(); n--)
    {
        std::ostrstream sout;
        sout << std::fixed << std::setw(sWidth) << std::setprecision(sPrecision) << words[n];
        // assert(sout.good());
        str = sout.str();
        if (dummyValue == (long)str.c_str())
            break;
    }

    return str.c_str();
}

const char* TestStringStream(Words& words, long dummyValue)
{
    unsigned long sum = 1;
    static std::string str;
    for (size_t n = words.size()-1; sum != 0 && n < words.size(); n--)
    {
        std::ostringstream sout;
        sout << std::fixed << std::setw(sWidth) << std::setprecision(sPrecision) << words[n];
        // assert(sout.good());
        str = sout.str();
        sum += (unsigned long)str.c_str();
    }

    return str.c_str();
}

int main(int argc, char* argv[])
{
    Words floatWords;
    long dummyValue = argc;

    MakeFloatWords(1000, floatWords);
    TimeIt("fcvt", TestFcvt, floatWords, dummyValue);
    TimeIt("gcvt", TestGcvt, floatWords, dummyValue);
 //   TimeIt("ecvt", TestEcvt, floatWords, dummyValue);
    TimeIt("sprintf", TestSprintf, floatWords, dummyValue);
    TimeIt("strStream", TestStrStream, floatWords, dummyValue);
    TimeIt("stringStream", TestStringStream, floatWords, dummyValue);

	return 0;
}
